/*
 *	surrect.cpp
 *      CopyRight (C) 2007, Nikolaj Krivchenkov aka unDEFER <undefer@gmail.com>
 */

#define _LARGEFILE64_SOURCE

#include "any.h"
#include "super.h"

#if defined(_WIN32)
#include <windows.h>
#include <winioctl.h>

#define O_LARGEFILE	0
#endif

#include <sys/types.h>
#if defined(__linux__)
#include <sys/wait.h>
#define	MKDIR(a, b) mkdir(a, b)
#else
#include "misc.h"
#define	MKDIR(a, b) mkdir(a)
#define	ftruncate64(a, b) ftruncate(a, b)
#endif
#include <unistd.h>

#include <stdio.h>
#if defined(__linux__)
#include <stdio_ext.h>
#endif
#include <stdlib.h>
#include <string.h>
#include <stdarg.h>

#if defined(__linux__)
#include <sys/ioctl.h>
#include <linux/fd.h>

#include <dlfcn.h>
#endif

#include <assert.h>

#include <bitops.h>
#if defined(__linux__)
#include <endian.h>
#else
#include <sys/param.h>
#include "ext2_bitops.h"
#endif
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "anysurrect.h"

#include "new_inode.h"
#include "block_map.h"
#include "progress.h"
#include "version.h"

#include "mainform.h"

#include <QFileIconProvider>
#include <QMessageBox>
#include <locale.h>

#if defined(__linux__)
    
typedef void* dl_handle_t;
#define DLOPEN(dll) dlopen(dll, RTLD_LAZY | RTLD_GLOBAL)
#define DLSYM(handle, sym) dlsym(handle, sym)
#define DLSUF ".so"
#define DLCLOSE(handle) dlclose(handle)
    
#elif defined(_WIN32)

typedef HMODULE dl_handle_t;
#define DLOPEN(dll) LoadLibrary(dll)
#define DLSYM(handle, sym) GetProcAddress((HINSTANCE__*) handle, sym)
#define DLSUF ".dll"
#define DLCLOSE(handle) FreeLibrary(handle)
#define dlerror()	"UNKNOWN ERROR"

#endif


#define BLKGETSIZE _IO(0x12,96) /* return device size /512 (long *arg) */

extern "C"
{
const char* default_list_types =
#include "../surrectlist.conf"
;

const char* ext2fs_info_list_types =
"filesystem_info_ext2fs_direct_blocks_links "
"filesystem_info_ext2fs_indirect_blocks_links "
"filesystem_info_ext2fs_double_indirect_blocks_links "
"filesystem_info_ext2fs_inode_table "
"filesystem_info_ext2fs_group_info "
;

char* list_types = NULL;
char* list_types2 = NULL;

const char * program_name = "anysurrect";

int fd = -1;
char * libs = "";
int num_libs;
dl_handle_t *libs_handles;

struct progress_struct progress;

int list_all_types = 0;
int verbose = 0;
int quiet = 0;
int fromfirst = 0;
int mod_usages = 0;

const char *pathprefix = "";
const char *device_name;
const char *inode_table;
const char *input_inode_table = NULL;
unsigned long input_blocksize = 0;

any_size_t device_blocks = 0;

mode_t dir_umask = 0002;
mode_t file_umask = 0002;

const mode_t default_mode = 0666;
const int default_text = 0;
const char *default_indicator = "UNKNOWN";

typedef char *surrect_function_t();
typedef void parseopts_function_t(const int argc, const char* argv[]);
typedef void usage_function_t();

surrect_function_t** surrects;
const mode_t **modes;
const int **texts;
const char ***indicators;
char **typelines;

const char ***modoptions;
parseopts_function_t** parseopts;
usage_function_t** modusages;

const char ***modoptions2;
parseopts_function_t** parseopts2;
usage_function_t** modusages2;

static int ind_type=-1;
int num_types = 0; 


int num_types2 = 0; 
surrect_function_t** surrects2;
const mode_t **modes2;

static void usage(void)
{
	fprintf(stderr, _(
"Usage: %s [-b blocksize] [-i input_inode_table] [-p path_prefix]\n"
"\t[-u file_umask] [-U dir_umask] [-fqvV] [-g plug-ins]\n"
"\t[-e] [-t list_of_types] [-T list_of_types] [-l] device inode_table\n"),
			program_name);
	exit(1);
}

extern char *concat_strings(int n, ...);

unsigned long *block_bitmap;
struct frags_list *frags_list;
struct frags_list *file_template_frags_list;
struct frags_list *file_frags_list;

unsigned long blocks_before_frag;
struct frags_list *cur_frag;

struct io_buffer bufs[1024];

/*
int redirect_frags(struct frags_list *frags_list, unsigned long after,
		unsigned long block);
 */

uint32_t blocksize;
uint64_t blocksize64;
int 	 log2blocksize;
uint32_t bitsblocksize;
uint64_t bitsblocksize64;
uint32_t nbitsblocksize;
uint64_t nbitsblocksize64;

struct io_buffer io_buffer = {NULL, -1, 0};

struct frags_list *copy_file_template_frags_list = NULL;

inline int set_block(any_ssize_t s_block)
{
	cut_frags(&file_template_frags_list, 0, s_block);
	if (copy_file_template_frags_list)
		cut_frags(&copy_file_template_frags_list, 0, s_block);
	io_buffer.size = 0;

	/*
	printf ("%d, %d\n", file_template_frags_list->frag.fr_start,
			file_template_frags_list->frag.fr_length);
*/
	return (file_template_frags_list)?0:1;
}

any_off_t cur_offset = 0;

any_off_t _fd_seek(any_off_t offset, int whence)
{
	if (whence==SEEK_CUR)
		offset += cur_offset;
	
	if (whence==SEEK_END)
		offset += fd_size();

	if ( ( blocks_before_frag << get_log2blocksize() ) > offset || !cur_frag )
	{
		blocks_before_frag = 0;
		cur_frag = file_frags_list;
	}

	while ( ( ( blocks_before_frag + cur_frag->frag.fr_length ) <<
					get_log2blocksize() )<=offset )
	{
		blocks_before_frag += cur_frag->frag.fr_length;
		NEXT_FRAG(cur_frag);
		if ( !cur_frag )
		{
			if ( ( blocks_before_frag << get_log2blocksize() )<offset )
				return -1;
			lseek64(fd, 0, SEEK_END);
			cur_offset = offset;
			return cur_offset;
		}
	}

	lseek64(fd, ( ((any_size_t) cur_frag->frag.fr_start - blocks_before_frag) << 
				get_log2blocksize()) + offset, 
			SEEK_SET);

	cur_offset = offset;
	
	return cur_offset;
}

void anysurrect_frags_list(struct frags_list *l_file_frags_list,
		any_size_t size, const char *dmes);
}

MainWindow *mainwindow = NULL;

int q_mkpathino(char *path, uint32_t root, struct any_sb_info *info,
		uint32_t *ino)
{
	int res;
	struct any_inode *root_inode;
	struct any_dirent **pdirent;
	struct any_dirent *newdirent;
	uint32_t newino;
	char *slash;
	QTreeWidgetItem *item;
	QTreeWidgetItem *parent;
	
	if ( !test_bit(root, info->si_inode_bitmap) ) return -1;
		
        root_inode = info->si_inode_table + root;
	
	if ( !S_ISDIR(root_inode->i_mode) ) return -1;

	while ( (*path)=='/' ) path++;

	slash = strchr(path, '/');
	if (slash) slash[0] = '\0';
	
	pdirent = &root_inode->i_info.dir->d_dirent;

	for (; *pdirent; pdirent=&(*pdirent)->d_next)
	{
		if (strcmp((*pdirent)->d_name, path)==0)
			goto created;
	}
	
	res = any_new_inode(info, S_IFDIR | (~dir_umask & 0777), 
			NULL, root, &newino);
	if (res) return res;

	/*
	 * We need to update the inode link, so as it maybe
	 * changed with realloc in any_new_inode
	 */
        root_inode = info->si_inode_table + root;

	newdirent = (any_dirent*) calloc(sizeof(struct any_dirent), 1);
	if (!newdirent)
	{
		fprintf(stderr, _("Not enough memory\n"));
		exit(1);
	}

	root_inode->i_info.dir->d_ndirents++;
	(*pdirent) = newdirent;
	(*pdirent)->d_name = strdup(path);
	(*pdirent)->d_inode = newino;

	mainwindow->EmitAddItem(info, strdup(path), newino, root);
	/*item = mainwindow->GetItem(info, path, newino);
	parent = (QTreeWidgetItem *)
		root_inode->i_info.dir->d_data;

	parent->addChild(item);*/

created:
	if (slash)
	{
		slash[0] = '/';
		return q_mkpathino(slash, 
				(*pdirent)->d_inode, 
				info, ino);
	}
	else 
		if (ino) *ino = (*pdirent)->d_inode;

	return 0;
}

void anysurrect_file(struct any_sb_info *info, char *dirn, mode_t mode)
{
	int res;
	struct any_inode *dir_inode;
	struct any_inode *inode;
	struct any_dirent **pdirent;
	struct any_dirent *newdirent;
	uint32_t newino;
	uint32_t dirino;
	char name_buffer[1024];

	QTreeWidgetItem *item;
	QTreeWidgetItem *parent;


	char *dir = concat_strings(2, pathprefix, dirn);
	
	res = q_mkpathino(dir, 1, info, &dirino);

	free(dir);
	
	res = any_new_inode(info, S_IFREG | ( mode & (~file_umask & 0777)  ), 
			NULL, dirino, &newino);
	if (res) 
	{
		fprintf (stderr, _("error while creating inode\n"));
		exit(1);
	}

	res = snprintf (name_buffer, 1024, "inode_%d_block_%ld",
			newino, get_block() );
	if (res<0)
	{
		fprintf(stderr, _("error while format filename\n"));
		exit(1);
	}

	newdirent = (struct any_dirent *) calloc(sizeof(struct any_dirent), 1);
	if (!newdirent)
	{
		fprintf(stderr, _("Not enough memory\n"));
		exit(1);
	}

        dir_inode = info->si_inode_table + dirino;

	pdirent = &dir_inode->i_info.dir->d_dirent;

	for (; *pdirent; pdirent=&(*pdirent)->d_next);

	dir_inode->i_info.dir->d_ndirents++;
	(*pdirent) = newdirent;
	(*pdirent)->d_name = strdup(name_buffer);
	(*pdirent)->d_inode = newino;
	
	inode = info->si_inode_table + newino;
	inode->i_size = fd_seek(0, SEEK_CUR);
	
	pick_frags(&file_frags_list, 0,
			( (inode->i_size-1) >> get_log2blocksize()) + 1 );

/*	printf ("%d, %d\n", file_frags_list->frag.fr_start,
			file_frags_list->frag.fr_length);
*/	
	frags_list2array (file_frags_list, &inode->i_info.file_frags);
/*
	printf ("%d\n", inode->i_info.file_frags->fr_nfrags);
	printf ("%d, %d\n", inode->i_info.file_frags->fr_frags->fr_start,
			inode->i_info.file_frags->fr_frags->fr_length);
			*/

	mainwindow->EmitAddItem(info, strdup(name_buffer), newino, dirino);
	/*item = mainwindow->GetItem(info, name_buffer, newino);
	parent = (QTreeWidgetItem *)
		dir_inode->i_info.dir->d_data;
	assert(parent);
	parent->addChild(item);*/
	
}

struct any_sb_info *info_for_anysurrect_frags_list;

void anysurrect_frags_list(struct frags_list *l_file_frags_list,
		any_size_t size, const char *dmes)
{
	const char *mes;
	int type;
	struct frags_list *old_copy_file_template_frags_list =
		copy_file_template_frags_list;
	struct frags_list *old_file_template_frags_list =
		file_template_frags_list;
	struct frags_list *old_file_frags_list =
		file_frags_list;
	file_template_frags_list = l_file_frags_list;
	copy_file_template_frags_list = NULL;

	unsigned long old_blocks_before_frag = blocks_before_frag;
	struct frags_list *old_cur_frag = cur_frag;

	struct any_sb_info *info = info_for_anysurrect_frags_list;
	int goodsize = 0;

	io_buffer.size = 0;

	char *old_pathprefix = (char *) pathprefix;
	pathprefix = concat_strings(3, pathprefix, dmes, "/");

	for (type=0; type<num_types2; type++)
	{
		if (copy_file_template_frags_list)
		{
			file_frags_list = copy_file_template_frags_list;
			copy_file_template_frags_list = NULL;
		}
		else
			copy_frags_list (file_template_frags_list, &file_frags_list);

		blocks_before_frag = 0;
		cur_frag = file_frags_list;
		fd_seek(0, SEEK_SET);
		{
			mes = surrects2[type](); 
			if (mes) { 
				mode_t mode = *modes2[type];
				char* dupmes = strdup(mes);
				if (verbose) printf("file %s, block=%ld, size=%llx\n",
						mes, get_block(), fd_seek(0, SEEK_CUR));
				if ( fd_seek(0, SEEK_CUR) == size ) goodsize = 1;
				anysurrect_file(info, dupmes, mode);
				free(dupmes);
				free_frags_list (file_frags_list);
			}
			else {	
				copy_file_template_frags_list = file_frags_list;
			}
		}
	}

	if (size && !goodsize)
	{
		if (copy_file_template_frags_list)
		{
			file_frags_list = copy_file_template_frags_list;
			copy_file_template_frags_list = NULL;
		}
		else
			copy_frags_list (file_template_frags_list, &file_frags_list);

		blocks_before_frag = 0;
		cur_frag = file_frags_list;
		fd_seek(size, SEEK_SET);

		mode_t mode = *modes2[ind_type];
		char* dupmes = strdup("UNKNOWN");
		if (verbose) printf("file %s, block=%ld, size=%llx\n",
				mes, get_block(), fd_seek(0, SEEK_CUR));
		if ( fd_seek(0, SEEK_CUR) == size ) goodsize = 1;
		anysurrect_file(info, dupmes, mode);
		free(dupmes);
		free_frags_list (file_frags_list);
	}

	if (copy_file_template_frags_list)
		free_frags_list(copy_file_template_frags_list);

	free_frags_list(l_file_frags_list);

	blocks_before_frag = old_blocks_before_frag;
	cur_frag = old_cur_frag;

	file_frags_list = old_file_frags_list;
	file_template_frags_list = old_file_template_frags_list;
	copy_file_template_frags_list = old_copy_file_template_frags_list;

	free((void*)pathprefix);
	pathprefix = old_pathprefix;

	io_buffer.size = 0;
}

static unsigned long *SKIP_TO_BLOCK = NULL;

void anysurrect_fromblock(struct any_sb_info *info)
{
	char *mes;
	any_size_t max_size=1;

	int type;

	info_for_anysurrect_frags_list = info;

	if (!SKIP_TO_BLOCK) 
	{
		SKIP_TO_BLOCK = 
			(unsigned long *)
			calloc(num_types, sizeof(unsigned long) );
		
		if (!SKIP_TO_BLOCK)
		{
			fprintf (stderr, _("Not enough memory\n"));
			exit(1);
		}
	}

	for (type=0; type<num_types; type++)
	{
		ind_type = type;
		int text = *texts[type];

		if (copy_file_template_frags_list)
		{
			file_frags_list = copy_file_template_frags_list;
			copy_file_template_frags_list = NULL;
		}
		else
			copy_frags_list (file_template_frags_list, &file_frags_list);

		blocks_before_frag = 0;
		cur_frag = file_frags_list;
		fd_seek(0, SEEK_SET);
		if (!text || (get_block()>=SKIP_TO_BLOCK[type]) ) {
			mes = surrects[type](); 
			if (mes) { 
				mode_t mode = *modes[type];
				char* dupmes = strdup(mes);
				max_size = max_t(any_size_t, max_size, fd_seek(0, SEEK_CUR) );
				anysurrect_file(info, dupmes, mode);
				free(dupmes);
				free_frags_list (file_frags_list);
			}
			else {	
				any_size_t pos = fd_seek(0, SEEK_CUR);
				if (text) SKIP_TO_BLOCK[type] = get_block() + 
					((pos-1) >> get_log2blocksize()) + 1;
				copy_file_template_frags_list = file_frags_list;
			}
		}
		else copy_file_template_frags_list = file_frags_list;
	}

	ind_type = -1;

	set_block( ( (max_size-1) >> get_log2blocksize() ) + 1 );
}

void MainWindow::EmitAddItem(struct any_sb_info *info,            
		const char *name, uint32_t newino, uint32_t dirino)
{
	emit SigAddItem(info, name, newino, dirino);
}

void MainWindow::AddItem(struct any_sb_info *info,            
		const char *name, uint32_t newino, uint32_t dirino)
{
	if (!state) 
	{
		printf ("state = 0\n");
		return; 
	}
	QTreeWidgetItem *item;
	QTreeWidgetItem *parent;
	struct any_inode *dir_inode;
        
	item = mainwindow->GetItem(info, name, newino);
	dir_inode = info->si_inode_table + dirino;
	parent = (QTreeWidgetItem *)
		dir_inode->i_info.dir->d_data;
	assert(parent);
	parent->addChild(item);

	free( (void*) name );
}

int MainWindow::FillItems(uint32_t ino, char *name, QTreeWidgetItem *parent)
{
	struct any_sb_info *info = sb_info;

	struct any_dirent* dirent;

	QTreeWidgetItem *item = GetItem(info, name, ino);
	if (!parent)
		treeWidget->insertTopLevelItem(0, item);
	else
		parent->addChild(item);

	struct any_inode *inode = info->si_inode_table + ino;
	if ( inode->i_mode>>12 == 04 )
	{
		dirent = info->si_inode_table[ino].i_info.dir->d_dirent;

		for(; dirent; dirent = dirent->d_next)
		{
			FillItems(dirent->d_inode, dirent->d_name, item);
			if (!open_state) return 1;
		}
	}

	QCoreApplication::processEvents();

	if (!open_state) return 1;

	return 0;
}

QTreeWidgetItem *MainWindow::GetItem(struct any_sb_info *info,
		const char *name, uint32_t ino)
{
	QFileIconProvider ip;
	int icontype = -1;

	QStringList columnData;
	struct any_inode *inode = info->si_inode_table + ino;
	columnData << QString(name);
	QString str; 
	str.setNum(inode->i_size); columnData << str;

	switch(inode->i_mode>>12)
	{
		case 01:
			columnData << _("FIFO");
			break;
		case 02:
			columnData << _("Character Device");
			break;
		case 04:
			columnData << _("Directory");
			icontype = QFileIconProvider::Folder;
			break;
		case 06:
			columnData << _("Block Device");
			icontype = QFileIconProvider::Drive;
			break;
		case 010:
			columnData << _("Regular File");
			icontype = QFileIconProvider::File;
			break;
		case 012:
			columnData << _("Symbolic Link");
			break;
		case 014:
			columnData << _("Socket");
			break;
		default:
			columnData << _("Unknown");
			break;
	}

	if (ino==1)
		icontype = QFileIconProvider::Drive;

	if ( inode->i_mode>>12 == 012 )
		columnData << inode->i_info.symlink;
	else if ( inode->i_mode>>12 != 010 &&
	       inode->i_mode>>12 != 04 )
	{
		str.setNum(inode->i_info.device); 
		columnData << str;
	}
	else
		columnData << "N/A";

	str.setNum(inode->i_mode & 0xFFF, 8); columnData << str;
	str.setNum(inode->i_uid); columnData << str;
	str.setNum(inode->i_gid); columnData << str;
	char timestr[100];
	timestr[0]='\0'; strftime(timestr, 100, "%c",
			localtime((const time_t*)&inode->i_atime));
	columnData << QString::fromUtf8(timestr);
	timestr[0]='\0'; strftime(timestr, 100, "%c",
			localtime((const time_t*)&inode->i_ctime));
	columnData << QString::fromUtf8(timestr);
	timestr[0]='\0'; strftime(timestr, 100, "%c",
			localtime((const time_t*)&inode->i_mtime));
	columnData << QString::fromUtf8(timestr);
	str.setNum(inode->i_links_count); columnData << str;
	str.setNum(ino); columnData << str;

	QTreeWidgetItem *item = new QTreeWidgetItem(columnData);
	if (icontype>=0)
	item->setIcon( 0, ip.icon(QFileIconProvider::IconType(icontype)) );

	if ( inode->i_mode>>12 == 04 )
		inode->i_info.dir->d_data = (void *) item;

	return item;
}

int MainWindow::SaveInode(const QString &To,
		uint32_t ino)
{
	struct any_sb_info *info = sb_info;
	struct any_inode *inode = info->si_inode_table + ino;
	struct any_dirent* dirent;
	int res;
	int wfd, rfd;
	struct any_file_frags *file_frags;
	struct any_file_fragment *frag;
	int i, j;
	int bytes_for_write;
	char *buf;
	int result = 0;

	const char *name = strdup(To.toAscii().data());

	if ( !access(name, F_OK) )
	{
		if ( unlink(name) )
		{
			QMessageBox::critical(this, _("Cannot remove file"), 
					_("I need in write access to the file"));
			result=1; goto out;
		}
	}

	switch (inode->i_mode >> 12)
	{
		/*Is FIFO*/
		case 01:
		/*Is SymDevice*/
		case 02:
		/*Is Block Device*/
		case 06:
		/*Socket*/
		case 014:
#if defined(__linux__)
			if ( mknod(name, inode->i_mode, inode->i_info.device) )
			{
				QMessageBox::critical(this, _("Cannot create special file"), 
						_("I need in write access to the directory"));
				result=1; goto out;
			}
#endif
			break;
		/*Is Directory*/
		case 04:
			if ( MKDIR(name, inode->i_mode & 0777) )
			{
				QMessageBox::critical(this, _("Cannot make directory"), 
						_("I need in write access to the directory"));
				result=1; goto out;
			}

			dirent = inode->i_info.dir->d_dirent;
			for (; dirent; dirent=dirent->d_next)
			{
				QString NewFile = To + "/" + dirent->d_name;
				res = SaveInode(NewFile, dirent->d_inode);
				if (res) {result=res; goto out;}
			}
			break;
		/*Is Regular File*/
		case 010:
#if defined(__linux__)
			wfd = open(name, O_WRONLY | O_CREAT | O_LARGEFILE, 
					inode->i_mode & 0777);
#elif defined(_WIN32)
			wfd = open(name, O_WRONLY | O_CREAT | O_BINARY, 
					inode->i_mode & 0777);
#endif
			if (wfd<0)
			{
				fprintf(stderr,
						_("cannot open '%s' for write: %s\n"),
						name,
						(errno)?strerror(errno):_("write error"));
				QMessageBox::critical(this, _("Cannot create regular file"), 
						_("I need in write access to the directory"));
				result=1; goto out;
			}

#if defined(__linux__)
			rfd = open(device_name, O_RDONLY | O_LARGEFILE);
#elif defined(_WIN32)
			rfd = open(device_name, O_RDONLY | O_BINARY);
#endif
			if (rfd==-1) {
				QMessageBox::critical(this, _("Cannot open device"), 
						_("I need in read access to the device"));
				result=1; goto out;
			}

			buf = (char*) malloc(info->si_blocksize);
			if (!buf)
			{
				fprintf(stderr, "Cannot allocate memory\n");
				exit(1);
			}

			file_frags = inode->i_info.file_frags;

			for (i=0; i < file_frags->fr_nfrags; i++)
			{
				frag = file_frags->fr_frags + i;

				lseek64(rfd, (any_size_t) get_blocksize64() * 
						frag->fr_start, SEEK_SET);
				for (j=0; j < frag->fr_length; j++)
				{
					bytes_for_write =
						min_t(int,
						get_blocksize(),
						inode->i_size - lseek64(wfd, 0, SEEK_CUR));

					res = READ (rfd, buf, get_blocksize());
					if (res < 0)
					{
						QMessageBox::critical(this, _("Cannot read from device"), 
								_("Error in inode table?"));
						perror("Cannot read from device");
						result=1; goto FREE;
					}

					res = write (wfd, buf, bytes_for_write);
					if (res < 0)
					{
						QMessageBox::critical(this, _("Cannot write to the file"), 
								_("Not enough space?"));
						perror("Cannot write to the file");
						result=1; goto FREE;
					}
				}
			}

FREE:
			free(buf);
			CLOSE(rfd);
			::close(wfd);
			break;
		/*Is SymLink*/
		case 012:
#if defined(__linux__)
			if ( symlink(name, inode->i_info.symlink) )
			{
				QMessageBox::critical(this, _("Cannot create symbolic link"), 
						_("I need in write access to the directory"));
				result=1; goto out;
			}
#endif
			break;
		default:
			break;
	}

out:
	free((void*) name);
	return result;
}

int MainWindow::SaveSelected()
{
	device_name = strdup(diskBox->currentText().toAscii().data());

	QTreeWidgetItem *item = treeWidget->currentItem();
	uint32_t ino = item->text(11).toInt();

	QFileDialog fd(this);

	QString FileName = fd.getSaveFileName();
	if ( FileName.isNull() )
		return 1;

	return SaveInode(FileName, ino);
}

int MainWindow::InitSurrect()
{
	char *begin;
	char *end;

	int i;
	for (i=0;i<1024;i++) bufs[i].buffer = NULL;

	mainwindow = this;

	libs = (char*) strdup(libs);
	list_types = (char*) strdup(default_list_types);
	list_types2 = (char*) strdup(default_list_types);

	num_libs = 1;
	begin = libs;
	while ( begin[0] == ' ' ) begin++;
	while ( (end = strchr(begin, ' ')) || ( begin[0] && 
				(end = begin+strlen(begin)) ) )
	{
		num_libs++;
		begin = end;
		while ( begin[0] == ' ' ) begin++;
	}
	
	libs_handles = (dl_handle_t *) 
		MALLOC( sizeof(dl_handle_t) * num_libs );

#if defined(__linux__)
	libs_handles[0] = DLOPEN(NULL);
#elif defined(_WIN32)
	libs_handles[0] = GetModuleHandle(NULL);
#endif
	if (!libs_handles[0])
	{
		fprintf( stderr, "Error dlopen() main handle: %s\n",
				dlerror() );
		exit(1);
	}
	
	int lib = 1;
	begin = libs;
	while ( begin[0] == ' ' ) begin++;
	while ( (end = strchr(begin, ' ')) || ( begin[0] && 
				(end = begin+strlen(begin)) ) )
	{
		char old = end[0];
		end[0] = '\0';

		libs_handles[lib] = DLOPEN(begin);
		if (!libs_handles[lib])
		{
			fprintf( stderr, "Error dlopen() %s: %s\n",
					begin, dlerror() );
			exit(1);
		}

		end[0] = old;
		lib++;

		begin = end;
		while ( begin[0] == ' ' ) begin++;
	}

	begin = list_types;
	while ( begin[0] == ' ' ) begin++;
	while ( (end = strchr(begin, ' ')) || ( begin[0] && 
				(end = begin+strlen(begin)) ) )
	{
		num_types++;
		begin = end;
		while ( begin[0] == ' ' ) begin++;
	}

	begin = list_types2;
	while ( begin[0] == ' ' ) begin++;
	while ( (end = strchr(begin, ' ')) || ( begin[0] && 
				(end = begin+strlen(begin)) ) )
	{
		num_types2++;
		begin = end;
		while ( begin[0] == ' ' ) begin++;
	}
	
	surrects = (surrect_function_t**) 
		MALLOC( sizeof(surrect_function_t*) * num_types );
	modes = (const mode_t**)
		MALLOC( sizeof(mode_t*) * num_types );
	texts = (const int**)
		MALLOC( sizeof(int*) * num_types );
	indicators = (const char***)
		MALLOC( sizeof(char**) * num_types );

	modoptions = (const char***)
		MALLOC( sizeof(char**) * num_types );
	parseopts = (parseopts_function_t**)
		MALLOC( sizeof(parseopts_function_t*) * num_types );
	modusages = (usage_function_t**)
		MALLOC( sizeof(usage_function_t*) * num_types );

	surrects2 = (surrect_function_t**) 
		MALLOC( sizeof(surrect_function_t*) * num_types2 );
	modes2 = (const mode_t**)
		MALLOC( sizeof(mode_t*) * num_types2 );

	modoptions2 = (const char***)
		MALLOC( sizeof(char**) * num_types2 );
	parseopts2 = (parseopts_function_t**)
		MALLOC( sizeof(parseopts_function_t*) * num_types2 );
	modusages2 = (usage_function_t**)
		MALLOC( sizeof(usage_function_t*) * num_types2 );

	int type = 0;
	begin = list_types;
	while ( begin[0] == ' ' ) begin++;
	while ( (end = strchr(begin, ' ')) || ( begin[0] && 
				(end = begin+strlen(begin)) ) )
	{
		void *handle = NULL;
		int n;
		char buf[1024];

		char old = end[0];
		end[0] = '\0';

		n=snprintf(buf, 1024, "%s_surrect", begin);
		if (n<0) exit(1);

		for (lib=0; lib<num_libs; lib++)
		{
			surrects[type] = (surrect_function_t*)
				DLSYM(libs_handles[lib], buf);
			if (surrects[type])
			{
				handle = libs_handles[lib];
				break;
			}
		}

		if (!handle)
		{
			fprintf (stderr,
					_("can't find %s surrecter function\n"),
					begin);
			exit(1);
		}

		n=snprintf(buf, 1024, "%s_mode", begin);
		if (n<0) exit(1);

		modes[type] = (mode_t*) DLSYM(handle, buf);
		if (!modes[type])
			modes[type] = &default_mode;

		n=snprintf(buf, 1024, "%s_text", begin);
		if (n<0) exit(1);

		texts[type] = (int*) DLSYM(handle, buf);
		if (!texts[type])
			texts[type] = &default_text;

		n=snprintf(buf, 1024, "%s_indicator", begin);
		if (n<0) exit(1);

		indicators[type] = (const char**) DLSYM(handle, buf);
		if (!indicators[type])
			indicators[type] = &default_indicator;

		n=snprintf(buf, 1024, "%s_opts", begin);
		if (n<0) exit(1);

		modoptions[type] = (const char**) DLSYM(handle, buf);

		n=snprintf(buf, 1024, "%s_parseopts", begin);
		if (n<0) exit(1);

		parseopts[type] = (parseopts_function_t*) DLSYM(handle, buf);

		n=snprintf(buf, 1024, "%s_usage", begin);
		if (n<0) exit(1);

		modusages[type] = (usage_function_t*) DLSYM(handle, buf);

		end[0] = old;
		type++;

		begin = end;
		while ( begin[0] == ' ' ) begin++;
	}
	
	type = 0;
	begin = list_types2;
	while ( begin[0] == ' ' ) begin++;
	while ( (end = strchr(begin, ' ')) || ( begin[0] && 
				(end = begin+strlen(begin)) ) )
	{
		void *handle = NULL;
		int n;
		char buf[1024];

		char old = end[0];
		end[0] = '\0';

		n=snprintf(buf, 1024, "%s_surrect", begin);
		if (n<0) exit(1);

		for (lib=0; lib<num_libs; lib++)
		{
			surrects2[type] = (surrect_function_t*)
				DLSYM(libs_handles[lib], buf);
			if (surrects2[type])
			{
				handle = libs_handles[lib];
				break;
			}
		}

		if (!handle)
		{
			fprintf (stderr,
					_("can't find %s surrecter function\n"),
					begin);
			exit(1);
		}

		n=snprintf(buf, 1024, "%s_mode", begin);
		if (n<0) exit(1);

		modes2[type] = (mode_t*) DLSYM(handle, buf);
		if (!modes2[type])
			modes2[type] = &default_mode;

		int no_in_first_list = 1;
		int type1;

		for (type1=0; type1 < num_types; type1++)
		{
			if (surrects[type1] == surrects2[type])
			{
				no_in_first_list = 0;
				break;
			}
		}

		if (no_in_first_list)
		{
			n=snprintf(buf, 1024, "%s_opts", begin);
			if (n<0) exit(1);

			modoptions2[type] = (const char**) 
				DLSYM(handle, buf);

			n=snprintf(buf, 1024, "%s_parseopts", begin);
			if (n<0) exit(1);

			parseopts2[type] = (parseopts_function_t*) 
				DLSYM(handle, buf);

			n=snprintf(buf, 1024, "%s_usage", begin);
			if (n<0) exit(1);

			modusages2[type] = (usage_function_t*) 
				DLSYM(handle, buf);
		}
		else
		{
			modoptions2[type] = NULL;
			parseopts2[type] = NULL;
			modusages2[type] = NULL;
		}

		end[0] = old;
		type++;

		begin = end;
		while ( begin[0] == ' ' ) begin++;
	}

	if (mod_usages)
	{
		for (type=0; type < num_types; type++)
			if (modusages[type])
				modusages[type](); 

		for (type=0; type < num_types2; type++)
			if (modusages2[type])
				modusages2[type](); 

		exit(0);
	}

	typelines = (char**)
		MALLOC( sizeof(char*) * num_types );

	type = 0;

	for (type=0; type<num_types; type++)
	{
		typelines[type] = (char*) MALLOC(33);

		typelines[type][32] = '\0';
		typelines[type][0] = ' ';
		typelines[type][1] = '[';
		int typeline_length = strlen(*indicators[type]);
		typeline_length = min_t(int, typeline_length, 29);
		memcpy(typelines[type]+2, *indicators[type], typeline_length);
		typelines[type][2+typeline_length] = ']';
		memset (typelines[type] + (3+typeline_length), ' ', 
				32 - (3+typeline_length) );
	}

	qRegisterMetaType<QModelIndex>("QModelIndex");
}

int MainWindow::DeInitSurrect()
{
	int lib;

	for (lib=0; lib<num_libs; lib++)
	{
		if ( DLCLOSE(libs_handles[lib]) )
		{
			fprintf( stderr, "Error dlclose(): %s\n",
					dlerror() );
			exit(1);
		}
	}

	anysurrect_free_clean();
}

void MainWindow::Message(QString str1, QString str2)
{
	QMessageBox::critical(this, str1, str2);
}

int MainWindow::StartSurrect()
{
	int r;
	uint32_t rootino;
	struct any_sb_info *info;
	unsigned long bitmap_l;
	any_size_t blocks;
	int pr;

	fromfirst = firstblockBox->isChecked();

	state = 1;
	blocksizeBox->setEnabled(false);
	firstblockBox->setEnabled(false);

	actionNew->setEnabled(1);
	actionSave->setEnabled(1);
	actionSave_As->setEnabled(1);

	input_blocksize = blocksizeBox->currentText().toUInt();
	device_name = strdup(diskBox->currentText().toAscii().data());
	inode_table = "inode_table.it";

#if defined(__linux__)
	FD = fd = open(device_name, O_RDONLY | O_LARGEFILE);
#elif defined(_WIN32)
	FD = fd = open(device_name, O_RDONLY | O_LARGEFILE | O_BINARY);
#endif
	if (fd==-1) {
		fprintf(stderr, _("Error opening file\n"));
		emit SigMessage(_("Cannot open file"), 
				_("Error opening file"));
		return 1;
	}

	uint64_t size = 0;
#if defined(__linux__)
	struct stat64 st;
	fstat64(fd, &st);
	if (st.st_mode & S_IFBLK)
	{
		if (ioctl(fd, BLKGETSIZE, &st.st_size) < 0) {
			perror("BLKGETSIZE");
			CLOSE(fd);
			return -EIO;
		}
		st.st_size*=512;
	}

	size = st.st_size;
#elif defined(_WIN32)
	struct stat st;
	fstat(fd, &st);
	if ( strstr(device_name, "\\\\.\\")==device_name )
	{
		HANDLE hDevice;               // handle to the drive to be examined 
		BOOL bResult;                 // results flag
		DWORD junk;                   // discard results
		DISK_GEOMETRY pdg;
		PARTITION_INFORMATION_EX	pi;

		hDevice = CreateFile(device_name,  // drive 
				0,                // no access to the drive
				FILE_SHARE_READ | // share mode
				FILE_SHARE_WRITE, 
				NULL,             // default security attributes
				OPEN_EXISTING,    // disposition
				0,                // file attributes
				NULL);            // do not copy file attributes

		if (hDevice == INVALID_HANDLE_VALUE) // cannot open the drive
			return -EIO;

#if 0
		bResult = DeviceIoControl(hDevice,  // device to be queried
				FSCTL_ALLOW_EXTENDED_DASD_IO,  // operation to perform
				NULL, 0, // no input buffer
				NULL, 0, // output buffer
				&junk,                 // # bytes returned
				(LPOVERLAPPED) NULL);  // synchronous I/O
#endif

		bResult = DeviceIoControl(hDevice,  // device to be queried
				IOCTL_DISK_GET_PARTITION_INFO_EX,  // operation to perform
				NULL, 0, // no input buffer
				&pi, sizeof(pi),     // output buffer
				&junk,                 // # bytes returned
				(LPOVERLAPPED) NULL);  // synchronous I/O

		size = pi.PartitionLength.QuadPart;

		if (!bResult)
		{
			bResult = DeviceIoControl(hDevice,  // device to be queried
					IOCTL_DISK_GET_DRIVE_GEOMETRY,  // operation to perform
					NULL, 0, // no input buffer
					&pdg, sizeof(pdg),     // output buffer
					&junk,                 // # bytes returned
					(LPOVERLAPPED) NULL);  // synchronous I/O

			size = pdg.Cylinders.QuadPart * pdg.TracksPerCylinder *
				pdg.SectorsPerTrack * pdg.BytesPerSector;
		}

		CloseHandle(hDevice);
	}
	else
		size = st.st_size;
#endif

	if (!input_inode_table)
	{
		if (!input_blocksize)
		{
			set_blocksize(512);

			while ( size >> get_log2blocksize() > 0xFFFFFF )
				set_blocksize(get_blocksize()*2);

			if ( get_blocksize()>4096 )
				set_blocksize(4096);
		}
		else
			set_blocksize(input_blocksize);


		if ( size >> get_log2blocksize() > 0xFFFFFFFFULL )
		{
			emit SigMessage(_("Too large device"), 
					_("Sorry, too large device."));
			fprintf(stderr,
					_("Sorry, too large device.\n"));
			goto out;
		}

		uint32_t inodes = ( size >> get_log2blocksize() )/1024 + 256;

		r = alloc_it(&info, get_blocksize(), inodes);
		if (r<0) goto out;

		r = any_new_inode(info, S_IFDIR | (~dir_umask & 0777), NULL, 0, &rootino);
		if (r) goto free_out;

		if (rootino!=1)
		{
			fprintf (stderr, _("Error creating root inode\n"));
			exit(1);
		}

		QTreeWidgetItem *item = GetItem(info, "/", rootino);

		treeWidget->insertTopLevelItem(0, item);
	}
	else
	{
		r = read_it (&info, (char*) input_inode_table);
		if (r)
		{
			emit SigMessage(_("Error open inode table"), 
					_("Error while reading inode table"));
			fprintf(stderr,
					_("Error while reading inode table: %s\n"),
					(errno)?strerror(errno):_("bad format"));
			goto out;
		}
		
		if (input_blocksize && input_blocksize!=info->si_blocksize)
		{
			fprintf(stderr,
_("Specified input inode table has %lu blocksize,\n"
"  but specified blocksize by -b option is %lu\n"
"Use 'reblock' for changing blocksize of inode_table\n"),
	info->si_blocksize, input_blocksize);
			exit(1);
		}
		
		set_blocksize(info->si_blocksize);
	}

	if (SKIP_TO_BLOCK)
	{
		int type;
		for (type=0; type<num_types; type++)
		{
			int text = *texts[type];
			if (text) SKIP_TO_BLOCK[type] = 0;
		}
	}

	sb_info = info;
	
	blocks = ((size-1) >> get_log2blocksize()) + 1;
	device_blocks = blocks;

	if (!blocks)
	{
		emit SigMessage(_("Too small device"), 
				_("Too small device for this blocksize"));
		r=1; goto out;
	}

	bitmap_l = ( blocks +
			sizeof(unsigned long) - 1 )/sizeof(unsigned long);

	/*allocate memory for block bitmap*/
	block_bitmap =  (typeof(block_bitmap))
		calloc(bitmap_l*sizeof(unsigned long), 1);
	if (!block_bitmap)      
	{ 
		fprintf(stderr, _("Not enough memory\n"));
		exit(1); 
	}

	r = fill_block_bitmap (info, block_bitmap, blocks, 0);
	if (r) exit(r);

	if (fromfirst) clear_bit(0, block_bitmap);

	if (blocks <= 1 && !fromfirst)
	{
		emit SigMessage(_("Too small device"), 
				_("Too small device for this blocksize"));
		r=1; goto out;
	}

	anysurrect_malloc_clean();
	
	create_frags_list(block_bitmap, blocks, &frags_list);

	io_buffer.size = 0;
	io_buffer.buffer = (char*) malloc(get_blocksize());
	if (!io_buffer.buffer)
	{
		fprintf(stderr, _("Not enough memory\n"));
		exit(1);
	}
	
	copy_frags_list (frags_list, &file_template_frags_list);

	progressBar->setMaximum(blocks);
	emit SetProgress(0);

	pr = blocks/1000;
	if (pr<1) pr = 1;

	for ( ; !set_block(0) && !wait_terminate; )
	{
		if ( save_state == 1 ) save_state++;
		while ( save_state > 1 ) 
#if defined(__linux__)
			usleep(1000);
#else
			Sleep(1);
#endif

		if (! (file_template_frags_list->frag.fr_start%pr) )
		emit SetProgress(file_template_frags_list->frag.fr_start);

		anysurrect_fromblock(info);
	}

	if (copy_file_template_frags_list)
	{
		free_frags_list(copy_file_template_frags_list);
		copy_file_template_frags_list = NULL;
	}

	if (file_template_frags_list)
	{
		free_frags_list(file_template_frags_list);
		file_template_frags_list = NULL;
	}

	if (frags_list)
	{
		free_frags_list(frags_list);
		frags_list = NULL;
	}

	free(io_buffer.buffer);
	io_buffer.buffer = NULL;

	if (!wait_terminate)
	{
		emit SetProgress(blocks);
		emit SigFinishSurrect();
	}

free_out:
	//free_it (info);

out:

	return r;
}

void MainWindow::FinishSurrect()
{
	if (thread)
	{
		if ( !thread->isFinished() )
		{
			wait_terminate = 1;
			thread->wait();
			wait_terminate = 0;
		}

		delete thread;
		thread = NULL;
	}
	
	anysurrect_free_clean();
	if (block_bitmap) free(block_bitmap);
	block_bitmap = NULL;
	if (FD!=-1)
		CLOSE(FD);
	FD = -1;
	fd = FD;
	state = 2;
}

int MainWindow::NewSurrect()
{
	if ( state && (state==1 || !saved) )
	{
		if ( !maybeSave() )
			return 1;
	}

	FinishSurrect();
	if (sb_info)
		free_it (sb_info);
	sb_info = NULL;
	treeWidget->clear();
	state = 0;

	blocksizeBox->setEnabled(true);
	firstblockBox->setEnabled(true);

	actionNew->setEnabled(0);
	actionSave->setEnabled(0);
	actionSave_As->setEnabled(0);

	saved = 0;

	open_state = 0;

	return 0;
}

int MainWindow::save()
{
	int r = 0;

	if ( state == 1 && !thread->isFinished() ) 
	{
		save_state = 1;
		while ( save_state < 2 && !thread->isFinished() ) 
#if defined(__linux__)
			usleep(1000);
#else
			Sleep(1);
#endif
	}

	r = write_it (sb_info, NULL);

	save_state = 0;

	if (r)
	{
		fprintf(stderr,
				_("Error while writing inode table: %s\n"),
				(errno)?strerror(errno):_("format error"));
		QMessageBox::critical(this, _("Write Error"), 
				_("Error while writing inode table."));
		return r;
	}

	if ( state == 2 )
	{
		saved = 1;
		actionSave->setEnabled(0);
		actionSave_As->setEnabled(0);
	}

	return 0;
}

int MainWindow::saveAction()
{
	if (!sb_info) return 1;
	if (!sb_info->si_itfilename)
	{
		QFileDialog fd(this);

		QString FileName = fd.getSaveFileName();
		if ( FileName.isNull() )
			return 1;

		sb_info->si_itfilename = strdup(FileName.toAscii().data());
	}

	return save();
}

int MainWindow::saveAsAction()
{
	if (!sb_info) return 1;
	QFileDialog fd(this);

	QString FileName = fd.getSaveFileName();
	if ( FileName.isNull() )
		return 1;

	sb_info->si_itfilename = strdup(FileName.toAscii().data());

	return save();
}

int MainWindow::openAction()
{
	int r;

	QFileDialog fd(this);
	QString FileName = fd.getOpenFileName();
	if ( FileName.isNull() )
		return 1;

	if ( NewSurrect() )
		return 1;

	char *input_inode_table = strdup(FileName.toAscii().data());

	r = read_it (&sb_info, (char*) input_inode_table);
	if (r)
	{
		fprintf(stderr,
				_("Error while reading inode table: %s\n"),
				(errno)?strerror(errno):_("bad format"));
		QMessageBox::critical(this, _("Open Error"), 
				_("Error while reading inode table."));
		return r;
	}

	set_blocksize(sb_info->si_blocksize);
	
	QString str;
	str.setNum(sb_info->si_blocksize);
	blocksizeBox->setEditText(str);

	blocksizeBox->setEnabled(false);
	firstblockBox->setEnabled(false);

	state = 2;
	saved = 1;

	actionNew->setEnabled(1);
	actionSave->setEnabled(0);
	actionSave_As->setEnabled(0);

	checkState();

	open_state = 1;

	FillItems();
}

int MainWindow::newAction()
{
	if ( ! NewSurrect() )
		checkState();
}

